from typing import TYPE_CHECKING

import pytest

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import (
    ComputePointsRuleNotDefinedError,
    compute_score_board,
    create_admin,
)
from yak_server.helpers.rules import Rules
from yak_server.v1.helpers.errors import NoAdminUser

if TYPE_CHECKING:
    from sqlalchemy import Engine


def test_score_board_rule_not_defined(
    monkeypatch: pytest.MonkeyPatch, engine_for_test_with_delete: "Engine"
) -> None:
    monkeypatch.setattr("yak_server.cli.database.get_settings", MockSettings(rules=Rules()))

    password_admin = get_random_string(10)

    create_admin(password_admin, engine_for_test_with_delete)

    with pytest.raises(ComputePointsRuleNotDefinedError) as exception:
        compute_score_board(engine_for_test_with_delete)

    assert str(exception.value) == "Compute points rule is not defined."


def test_score_board_rule_admin_user_not_found(engine_for_test_with_delete: "Engine") -> None:
    with pytest.raises(NoAdminUser) as exception:
        compute_score_board(engine_for_test_with_delete)

    assert str(exception.value) == "401: No admin user found"



from typing import Annotated

from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import UserModel
from yak_server.helpers.database import get_db
from yak_server.helpers.rules import RULE_MAPPING
from yak_server.helpers.settings import Settings, get_settings
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.errors import RuleNotFound, UnauthorizedAccessToAdminAPI
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut

router = APIRouter(prefix="/rules", tags=["rules"])


@router.post(
    "/{rule_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_ENTITY: {"model": ValidationErrorOut},
    },
)
def execute_rule(
    rule_id: UUID4,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(get_current_user)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> GenericOut[str]:
    rule_metadata = RULE_MAPPING.get(rule_id)

    if rule_metadata is None:
        raise RuleNotFound(rule_id)

    if rule_metadata.required_admin is True and user.name != "admin":
        raise UnauthorizedAccessToAdminAPI

    rule_metadata.function(db, user, getattr(settings.rules, rule_metadata.attribute))

    return GenericOut(result="")



from typing import TYPE_CHECKING

from fastapi import status
from starlette.testclient import TestClient

if TYPE_CHECKING:
    from fastapi import FastAPI


def is_client_error(code: int) -> bool:
    return 400 <= code < 500


def is_server_error(code: int) -> bool:
    return 500 <= code < 600


def test_check_response_models(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    openapi_schema_response = client.get("/api/openapi.json")

    assert openapi_schema_response.status_code == status.HTTP_200_OK

    openapi_schema = openapi_schema_response.json()

    for path, item in openapi_schema["paths"].items():
        if path == "/api/health/":
            assert item["get"]["tags"] == ["health"]

            assert "200" in item["get"]["responses"]

            assert "503" in item["get"]["responses"]
            assert (
                item["get"]["responses"]["503"]["content"]["application/json"]["schema"]["$ref"]
                == "#/components/schemas/ErrorOut"
            )
            assert item["get"]["responses"]["503"]["description"] == "Service Unavailable"

        if path.startswith("/api/v1"):
            for specification in item.values():
                assert (
                    str(status.HTTP_200_OK) in specification["responses"]
                    or str(status.HTTP_201_CREATED) in specification["responses"]
                )

                assert str(status.HTTP_422_UNPROCESSABLE_ENTITY) in specification["responses"]

                for status_code, schema in specification["responses"].items():
                    if int(status_code) == status.HTTP_422_UNPROCESSABLE_ENTITY:
                        assert (
                            schema["content"]["application/json"]["schema"]["$ref"]
                            == "#/components/schemas/ValidationErrorOut"
                        )
                    elif is_client_error(int(status_code)) or is_server_error(int(status_code)):
                        assert (
                            schema["content"]["application/json"]["schema"]["$ref"]
                            == "#/components/schemas/ErrorOut"
                        )



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_bets(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_bet_v2"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(10)
    first_name = get_random_string(5)
    last_name = get_random_string(8)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED
    authentication_token = response_signup.json()["result"]["access_token"]

    # Get all bets to retrieve ids
    response_get_all_bets = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_get_all_bets.status_code == HTTPStatus.OK

    score_bet_ids = [
        score_bet["id"] for score_bet in response_get_all_bets.json()["result"]["score_bets"]
    ]

    # Success case : check get one bet
    response_bet_by_id = client.get(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_bet_by_id.status_code == HTTPStatus.OK
    assert response_bet_by_id.json() == {
        "ok": True,
        "result": {
            "group": {
                "code": "A",
                "description": "Groupe A",
                "id": ANY,
            },
            "phase": {
                "code": "GROUP",
                "description": "Phase de groupes",
                "id": ANY,
            },
            "score_bet": {
                "id": ANY,
                "locked": False,
                "team1": {
                    "code": "HR",
                    "description": "Croatie",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
                "team2": {
                    "code": "FI",
                    "description": "Finlande",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
            },
        },
    }

    # Error case : bet with invalid id
    invalid_bet_id = uuid4()

    response_bet_with_invalid_id = client.get(
        f"/api/v1/score_bets/{invalid_bet_id}",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_bet_with_invalid_id.status_code == HTTPStatus.NOT_FOUND
    assert response_bet_with_invalid_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Bet not found: {invalid_bet_id}",
    }

    # Success case : bet by phase
    response_by_phase = client.get(
        "/api/v1/bets/phases/GROUP",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_by_phase.status_code == HTTPStatus.OK
    assert response_by_phase.json() == {
        "ok": True,
        "result": {
            "binary_bets": [],
            "groups": [
                {
                    "code": "A",
                    "description": "Groupe A",
                    "id": ANY,
                },
            ],
            "phase": {
                "code": "GROUP",
                "description": "Phase de groupes",
                "id": ANY,
            },
            "score_bets": [
                {
                    "group": {"id": ANY},
                    "id": ANY,
                    "locked": False,
                    "team1": {
                        "code": "HR",
                        "description": "Croatie",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                    "team2": {
                        "code": "FI",
                        "description": "Finlande",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                },
                {
                    "group": {"id": ANY},
                    "id": ANY,
                    "locked": False,
                    "team1": {
                        "code": "NA",
                        "description": "Namibie",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                    "team2": {
                        "code": "IQ",
                        "description": "Irak",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                },
                {
                    "group": {"id": ANY},
                    "id": ANY,
                    "locked": False,
                    "team1": {
                        "code": "HR",
                        "description": "Croatie",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                    "team2": {
                        "code": "NA",
                        "description": "Namibie",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                },
                {
                    "group": {"id": ANY},
                    "id": ANY,
                    "locked": False,
                    "team1": {
                        "code": "FI",
                        "description": "Finlande",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                    "team2": {
                        "code": "IQ",
                        "description": "Irak",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                },
                {
                    "group": {"id": ANY},
                    "id": ANY,
                    "locked": False,
                    "team1": {
                        "code": "HR",
                        "description": "Croatie",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                    "team2": {
                        "code": "IQ",
                        "description": "Irak",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                },
                {
                    "group": {"id": ANY},
                    "id": ANY,
                    "locked": False,
                    "team1": {
                        "code": "FI",
                        "description": "Finlande",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                    "team2": {
                        "code": "NA",
                        "description": "Namibie",
                        "flag": {"url": ANY},
                        "id": ANY,
                        "score": None,
                    },
                },
            ],
        },
    }

    # Error case : bet by phase invalid code
    invalid_phase_code = "GR"

    response_by_phase_with_invalid_code = client.get(
        f"/api/v1/bets/phases/{invalid_phase_code}",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_by_phase_with_invalid_code.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Phase not found: {invalid_phase_code}",
    }



import json
from pathlib import Path
from secrets import randbelow
from uuid import uuid4

import pytest
from click.testing import CliRunner

from testing.util import get_random_string
from yak_server.cli import app
from yak_server.cli.env import RuleNotDefinedError

runner = CliRunner()


def test_rule_not_defined(monkeypatch: pytest.MonkeyPatch) -> None:
    with runner.isolated_filesystem():
        competition_dir = Path("data", "fake_competition")

        rules_dir = competition_dir / "rules"
        rules_dir.mkdir(parents=True)

        rule_id = uuid4()

        (rules_dir / f"{rule_id}.json").write_text(json.dumps({}))

        (competition_dir / "config.ini").write_text(
            "[locking]\ndatetime = 2022-11-20T17:00:00.000000+01:00\n"
        )

        monkeypatch.setattr("yak_server.cli.env.__file__", Path("cli", "env.py"))

        host = get_random_string(20)
        user_name = get_random_string(6)
        password = get_random_string(100)
        port = randbelow(10000)
        database = get_random_string(20)

        result = runner.invoke(
            app,
            ["env", "all"],
            input=f"y\n{host}\n{user_name}\n{password}\n{port}\n{database}\n1800\n1800\nfake_competition\n",
        )

        assert result.exit_code == 1
        assert result.exception is not None
        assert isinstance(result.exception, RuleNotDefinedError)
        assert str(result.exception) == f"Rule not defined: {rule_id}"



from enum import Enum
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from yak_server.database.models import GroupModel, PhaseModel, TeamModel


class Lang(str, Enum):
    fr = "fr"
    en = "en"


def get_language_description(
    instance: Union["GroupModel", "PhaseModel", "TeamModel"],
    lang: Lang,
) -> str:
    if lang == Lang.fr:
        return instance.description_fr

    return instance.description_en


DEFAULT_LANGUAGE = Lang.fr



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

import pendulum
from starlette.testclient import TestClient

from testing.mock import MockLockDatetime, MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database
from yak_server.helpers.settings import get_lock_datetime

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_binary_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_binary_bet"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(2),
            "first_name": get_random_string(6),
            "last_name": get_random_string(12),
            "password": get_random_string(10),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    access_token = response_signup.json()["result"]["access_token"]

    response_bets = client.get("/api/v1/bets", headers={"Authorization": f"Bearer {access_token}"})

    assert response_bets.status_code == HTTPStatus.OK

    bet_id = response_bets.json()["result"]["binary_bets"][0]["id"]

    response_modify_binary_bet = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"is_one_won": True},
    )

    assert response_modify_binary_bet.status_code == HTTPStatus.OK
    assert response_modify_binary_bet.json() == {
        "ok": True,
        "result": {
            "phase": {"code": "GROUP", "description": "Phase de groupes", "id": ANY},
            "group": {"code": "A", "description": "Groupe A", "id": ANY},
            "binary_bet": {
                "id": ANY,
                "locked": False,
                "team1": {
                    "code": "FR",
                    "description": "France",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "won": True,
                },
                "team2": {
                    "code": "GR",
                    "description": "Allemagne",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "won": False,
                },
            },
        },
    }

    # Error case : locked bet
    app_with_valid_jwt_config.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") - pendulum.duration(minutes=10)
    )

    response_lock_bet = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"is_one_won": True},
    )

    assert response_lock_bet.status_code == HTTPStatus.UNAUTHORIZED
    assert response_lock_bet.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Cannot modify binary bet, lock date is exceeded",
    }

    app_with_valid_jwt_config.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    # Error case : Invalid input
    response_invalid_inputs = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"is_won": True},
    )

    assert response_invalid_inputs.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    # Error case : invalid bet id
    invalid_bet_id = uuid4()

    response_with_invalid_bet_id = client.patch(
        f"/api/v1/binary_bets/{invalid_bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"is_one_won": True},
    )

    assert response_with_invalid_bet_id.status_code == HTTPStatus.NOT_FOUND
    assert response_with_invalid_bet_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Bet not found: {invalid_bet_id}",
    }

    # Success case : retrieve one binary bet
    response_binary_bet_by_id = client.get(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_binary_bet_by_id.status_code == HTTPStatus.OK
    assert response_binary_bet_by_id.json() == {
        "ok": True,
        "result": {
            "phase": {"code": "GROUP", "description": "Phase de groupes", "id": ANY},
            "group": {"code": "A", "description": "Groupe A", "id": ANY},
            "binary_bet": {
                "id": ANY,
                "locked": False,
                "team1": {
                    "code": "FR",
                    "description": "France",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "won": True,
                },
                "team2": {
                    "code": "GR",
                    "description": "Allemagne",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "won": False,
                },
            },
        },
    }

    # Error case : retrieve binary bet with invalid id
    invalid_bet_id = uuid4()

    response_retrieve_with_invalid_bet_id = client.get(
        f"/api/v1/binary_bets/{invalid_bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_retrieve_with_invalid_bet_id.status_code == HTTPStatus.NOT_FOUND
    assert response_retrieve_with_invalid_bet_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Bet not found: {invalid_bet_id}",
    }

    # Success case : Modify team1 id by setting it to None
    response_change_team1_to_none = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"team1": {"id": None}},
    )

    assert response_change_team1_to_none.status_code == HTTPStatus.OK
    assert response_change_team1_to_none.json()["result"]["binary_bet"] == {
        "id": bet_id,
        "locked": False,
        "team1": None,
        "team2": {
            "id": ANY,
            "code": "GR",
            "description": "Allemagne",
            "flag": ANY,
            "won": False,
        },
    }

    # Success case : Change team2
    response_all_teams = client.get("/api/v1/teams")

    team_spain = [
        team for team in response_all_teams.json()["result"]["teams"] if team["code"] == "ES"
    ]

    assert len(team_spain) == 1

    team_spain = team_spain[0]["id"]

    response_change_team2 = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"team2": {"id": team_spain}},
    )

    assert response_change_team2.status_code == HTTPStatus.OK
    assert response_change_team2.json()["result"]["binary_bet"] == {
        "id": ANY,
        "locked": False,
        "team1": None,
        "team2": {
            "id": team_spain,
            "code": "ES",
            "description": "Espagne",
            "flag": ANY,
            "won": False,
        },
    }

    # Error case : modify team id with invalid team id
    invalid_team_id = str(uuid4())

    response_invalid_team1_id = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"team1": {"id": invalid_team_id}},
    )

    assert response_invalid_team1_id.status_code == HTTPStatus.NOT_FOUND
    assert response_invalid_team1_id.json()["description"] == f"Team not found: {invalid_team_id}"

    response_invalid_team2_id = client.patch(
        f"/api/v1/binary_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"team2": {"id": invalid_team_id}},
    )

    assert response_invalid_team2_id.status_code == HTTPStatus.NOT_FOUND
    assert response_invalid_team2_id.json()["description"] == f"Team not found: {invalid_team_id}"

    app_with_valid_jwt_config.dependency_overrides.clear()



import contextlib
import os
from collections.abc import Generator
from typing import TYPE_CHECKING

import pendulum
import psycopg
import pytest
from psycopg import sql
from sqlalchemy import Engine, create_engine

from scripts.profiling import create_app as create_app_with_profiling
from testing.mock import MockAuthenticationSettings, MockLockDatetime, MockSettings
from testing.util import get_random_string
from yak_server import create_app
from yak_server.cli.database import create_database, delete_database, drop_database
from yak_server.database import compute_database_uri, get_postgres_settings
from yak_server.helpers.rules import Rules
from yak_server.helpers.settings import get_authentication_settings, get_lock_datetime, get_settings

if TYPE_CHECKING:
    from fastapi import FastAPI


def create_test_database() -> Engine:
    db_settings = get_postgres_settings()

    with psycopg.connect(
        host=db_settings.host,
        user=db_settings.user,
        password=db_settings.password,
        port=db_settings.port,
        dbname="postgres",  # System database
    ) as connection:
        connection.autocommit = True  # Required to create a database

        with contextlib.suppress(psycopg.errors.DuplicateDatabase):
            connection.execute(sql.SQL("CREATE DATABASE {}").format(sql.Identifier(db_settings.db)))

    database_url = compute_database_uri(
        psycopg.__name__,
        db_settings.host,
        db_settings.user,
        db_settings.password,
        db_settings.port,
        db_settings.db,
    )

    return create_engine(database_url, pool_recycle=7200, pool_pre_ping=True)


@pytest.fixture(scope="session", autouse=True)
def engine_for_test() -> Generator[Engine, None, None]:
    engine = create_test_database()

    # Always drop database before running test session
    drop_database(engine, debug=True)

    # Always create database before running test session
    create_database(engine)

    yield engine

    # Always drop database after running test session
    drop_database(engine, debug=True)


@pytest.fixture
def engine_for_test_with_delete(engine_for_test: Engine) -> Generator[Engine, None, None]:
    delete_database(engine_for_test, debug=True)

    yield engine_for_test

    delete_database(engine_for_test, debug=True)


@pytest.fixture(scope="session")
def _debug_app_session() -> "FastAPI":
    os.environ["DEBUG"] = "1"

    return create_app()


@pytest.fixture(scope="session")
def production_app_session() -> "FastAPI":
    os.environ["DEBUG"] = "0"

    return create_app()


@pytest.fixture
def _app(
    _debug_app_session: "FastAPI", engine_for_test: "Engine"
) -> Generator["FastAPI", None, None]:
    # Clean database before running test
    delete_database(engine_for_test, debug=True)

    yield _debug_app_session

    # Clean database after running test
    delete_database(engine_for_test, debug=True)


@pytest.fixture
def _app_production(
    production_app_session: "FastAPI", engine_for_test_with_delete: "Engine"
) -> Generator["FastAPI", None, None]:
    # Clean database before running test
    delete_database(engine_for_test_with_delete, debug=True)

    yield production_app_session

    # Clean database after running test
    delete_database(engine_for_test_with_delete, debug=True)


@pytest.fixture
def app_with_profiler() -> Generator["FastAPI", None, None]:
    os.environ["DEBUG"] = "1"

    app = create_app_with_profiling()

    app.dependency_overrides[get_authentication_settings] = MockAuthenticationSettings(
        jwt_expiration_time=100,
        jwt_refresh_expiration_time=200,
        jwt_secret_key=get_random_string(15),
        jwt_refresh_secret_key=get_random_string(15),
    )

    app.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    yield app

    app.dependency_overrides.clear()


@pytest.fixture
def app_with_valid_jwt_config(_app: "FastAPI") -> Generator["FastAPI", None, None]:
    _app.dependency_overrides[get_authentication_settings] = MockAuthenticationSettings(
        jwt_expiration_time=100,
        jwt_refresh_expiration_time=200,
        jwt_secret_key=get_random_string(15),
        jwt_refresh_secret_key=get_random_string(15),
    )

    _app.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    yield _app

    _app.dependency_overrides.clear()


@pytest.fixture
def app_with_valid_jwt_config_production(
    _app_production: "FastAPI",
) -> Generator["FastAPI", None, None]:
    _app_production.dependency_overrides[get_authentication_settings] = MockAuthenticationSettings(
        jwt_expiration_time=100,
        jwt_refresh_expiration_time=200,
        jwt_secret_key=get_random_string(15),
        jwt_refresh_secret_key=get_random_string(15),
    )

    _app_production.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    yield _app_production

    _app_production.dependency_overrides.clear()


@pytest.fixture
def app_with_null_jwt_expiration_time(_app: "FastAPI") -> Generator["FastAPI", None, None]:
    _app.dependency_overrides[get_authentication_settings] = MockAuthenticationSettings(
        jwt_expiration_time=0,
        jwt_refresh_expiration_time=200,
        jwt_secret_key=get_random_string(15),
        jwt_refresh_secret_key=get_random_string(15),
    )

    _app.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    yield _app

    _app.dependency_overrides.clear()


@pytest.fixture
def app_with_null_jwt_refresh_expiration_time(_app: "FastAPI") -> Generator["FastAPI", None, None]:
    _app.dependency_overrides[get_authentication_settings] = MockAuthenticationSettings(
        jwt_expiration_time=100,
        jwt_refresh_expiration_time=3,
        jwt_secret_key=get_random_string(15),
        jwt_refresh_secret_key=get_random_string(15),
    )

    _app.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    yield _app

    _app.dependency_overrides.clear()


@pytest.fixture
def app_with_empty_rules(app_with_valid_jwt_config: "FastAPI") -> Generator["FastAPI", None, None]:
    app_with_valid_jwt_config.dependency_overrides[get_settings] = MockSettings(rules=Rules())

    yield app_with_valid_jwt_config

    app_with_valid_jwt_config.dependency_overrides.clear()



from http import HTTPStatus
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from fastapi.testclient import TestClient

from scripts.profiling import create_app as create_app_with_profiler
from testing.util import get_random_string

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_debug_profiling(app_with_profiler: "FastAPI") -> None:
    client = TestClient(app_with_profiler)

    user_name = get_random_string(6)

    # Check signup is ok with profiling
    response = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": get_random_string(10),
            "last_name": get_random_string(12),
            "password": get_random_string(18),
        },
    )

    assert response.status_code == HTTPStatus.CREATED
    assert "profiling-log-id" in response.headers

    file_name = response.headers["profiling-log-id"]

    profiling_file = Path(__file__).parents[1] / "profiling" / f"{file_name}.pstats"
    assert profiling_file.exists()


def test_profiling_without_yappi_installed(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("scripts.profiling.yappi", None)

    with pytest.raises(NotImplementedError) as exception:
        create_app_with_profiler()

    assert str(exception.value) == "Profiling is not available without yappi installed."



import json
from pathlib import Path
from secrets import randbelow

import pytest
from click.testing import CliRunner
from dotenv import dotenv_values

from testing.util import get_random_string
from yak_server.cli import app

runner = CliRunner()


def test_yak_env_all() -> None:
    host = get_random_string(10)
    user_name = get_random_string(15)
    password = get_random_string(250)
    port = randbelow(80000)
    database = get_random_string(15)

    with runner.isolated_filesystem():
        result = runner.invoke(
            app,
            ["env", "all"],
            input=f"y\n{host}\n{user_name}\n{password}\n{port}\n{database}\n1800\n1800\nworld_cup_2022\n",
        )

        assert result.exit_code == 0

        env = dotenv_values(Path(".env"))

        assert env["DEBUG"] == "1"
        assert env["JWT_EXPIRATION_TIME"] == "1800"
        assert env["JWT_SECRET_KEY"] is not None
        assert len(env["JWT_SECRET_KEY"]) == 256
        assert env["JWT_REFRESH_SECRET_KEY"] is not None
        assert len(env["JWT_REFRESH_SECRET_KEY"]) == 256
        assert env["COMPETITION"] == "world_cup_2022"
        assert env["LOCK_DATETIME"] == "2022-11-20T17:00:00+01:00"
        assert env["OFFICIAL_RESULTS_URL"] == "https://en.wikipedia.org/wiki/2022_FIFA_World_Cup"
        assert env["DATA_FOLDER"] is not None
        assert Path(env["DATA_FOLDER"]).samefile(
            Path(__file__).parents[1] / "yak_server" / "data" / "world_cup_2022"
        )
        assert env["RULES"] is not None
        assert len(json.loads(env["RULES"])) == 2

        env_db = dotenv_values(Path(".env.db"))

        assert env_db["POSTGRES_HOST"] == host
        assert env_db["POSTGRES_USER"] == user_name
        assert env_db["POSTGRES_PASSWORD"] == password
        assert env_db["POSTGRES_PORT"] == str(port)
        assert env_db["POSTGRES_DB"] == database


def test_yak_env_all_production() -> None:
    with runner.isolated_filesystem():
        result = runner.invoke(
            app,
            ["env", "all"],
            input="n\nlocalhost\nroot\ny\n3000\ndb\n1800\n1800\neuro_2016\n",
        )

        assert result.exit_code == 0

        env = dotenv_values(Path(".env"))

        assert env["DEBUG"] == "0"


def test_yak_env_all_world_cup_2018() -> None:
    with runner.isolated_filesystem():
        result = runner.invoke(
            app,
            ["env", "all"],
            input="y\nroot\nroot\ndddddddd\n\ndb\n1800\n1800\nworld_cup_2018\n",
        )

        assert result.exit_code == 0

        env = dotenv_values(Path(".env"))

        assert env["RULES"] is not None
        assert json.loads(env["RULES"]) == {}


def test_yak_env_all_invalid_lockdatetime(monkeypatch: pytest.MonkeyPatch) -> None:
    with runner.isolated_filesystem() as tmpdir:
        date = "2022-11-12"

        (Path(tmpdir) / "data" / "competition0").mkdir(parents=True)
        (Path(tmpdir) / "data" / "competition0" / "common.json").write_text(
            json.dumps({"lock_datetime": date, "official_results_url": ""})
        )

        monkeypatch.setattr("yak_server.cli.env.__file__", (Path(tmpdir) / "e" / "e").resolve())

        result = runner.invoke(
            app,
            ["env", "all"],
            input="n\nlocalhost\nroot\ny\n3000\ndb\n1800\n1800\ncompetition0\n",
            catch_exceptions=True,
        )

        assert result.exit_code == 1
        assert str(result.exception) == f"lock_datetime is not a valid datetime: {date}"


def test_yak_env_db() -> None:
    host = get_random_string(10)
    user_name = get_random_string(56)
    password = get_random_string(12)
    port = randbelow(50000)
    database = get_random_string(20)

    with runner.isolated_filesystem():
        result = runner.invoke(
            app, ["env", "db"], input=f"{host}\n{user_name}\n{password}\n{port}\n{database}\n"
        )

        assert result.exit_code == 0

        env_db = dotenv_values(Path(".env.db"))

        assert env_db["POSTGRES_HOST"] == host
        assert env_db["POSTGRES_USER"] == user_name
        assert env_db["POSTGRES_PASSWORD"] == password
        assert env_db["POSTGRES_PORT"] == str(port)
        assert env_db["POSTGRES_DB"] == database


def test_yak_env_app() -> None:
    with runner.isolated_filesystem():
        result = runner.invoke(
            app,
            ["env", "app"],
            input="y\n1800\n1800\nworld_cup_2018\n",
        )

        assert result.exit_code == 0

        env = dotenv_values(Path(".env"))

        assert env["DEBUG"] == "1"
        assert env["JWT_EXPIRATION_TIME"] == "1800"
        assert env["JWT_SECRET_KEY"] is not None
        assert len(env["JWT_SECRET_KEY"]) == 256
        assert env["JWT_REFRESH_SECRET_KEY"] is not None
        assert len(env["JWT_REFRESH_SECRET_KEY"]) == 256
        assert env["COMPETITION"] == "world_cup_2018"
        assert env["LOCK_DATETIME"] == "2018-06-14T18:00:00+03:00"
        assert env["OFFICIAL_RESULTS_URL"] == "https://en.wikipedia.org/wiki/2018_FIFA_World_Cup"
        assert env["DATA_FOLDER"] is not None
        assert Path(env["DATA_FOLDER"]).samefile(
            Path(__file__).parents[1] / "yak_server" / "data" / "world_cup_2018"
        )
        assert env["RULES"] is not None
        assert len(json.loads(env["RULES"])) == 0



from typing import Annotated

from fastapi import APIRouter, Depends, status
from sqlalchemy import text
from sqlalchemy.orm import Session

from .helpers.database import get_db
from .v1.models.generic import ErrorOut, GenericOut

router = APIRouter(prefix="/health", tags=["health"])


@router.get(
    "/",
    responses={status.HTTP_503_SERVICE_UNAVAILABLE: {"model": ErrorOut}},
    response_model_exclude_none=True,
)
def health_check(db: Annotated[Session, Depends(get_db)]) -> GenericOut[None]:
    db.execute(text("SELECT 1"))

    return GenericOut(ok=True, result=None)



from typing import TYPE_CHECKING

from pydantic import UUID4, BaseModel

from yak_server.helpers.language import Lang, get_language_description

if TYPE_CHECKING:
    from yak_server.database.models import PhaseModel


class PhaseOut(BaseModel):
    id: UUID4
    code: str
    description: str

    @classmethod
    def from_instance(cls, phase: "PhaseModel", *, lang: Lang) -> "PhaseOut":
        return cls(id=phase.id, code=phase.code, description=get_language_description(phase, lang))



from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any

from hatchling.builders.config import BuilderConfigBound
from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CopyAlembicIniBuildHook(BuildHookInterface[BuilderConfigBound]):
    PLUGIN_NAME = "copy-alembic-ini"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:  # noqa: ARG002
        root = Path(self.root)

        shutil.copy(root / "alembic.ini", root / "yak_server")

    def finalize(self, version: str, build_data: dict[str, Any], artifact_path: str) -> None:  # noqa: ARG002
        root = Path(self.root)

        (root / "yak_server" / "alembic.ini").unlink(missing_ok=True)



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_group_rank(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_compute_points_v1"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(10),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    access_token = response_signup.json()["result"]["access_token"]

    response_all_bets = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_all_bets.status_code == HTTPStatus.OK

    new_scores = [(5, 1), (0, 0), (1, 2)]

    for bet, new_score in zip(response_all_bets.json()["result"]["score_bets"], new_scores):
        response_patch_bet = client.patch(
            f"/api/v1/score_bets/{bet['id']}",
            headers={"Authorization": f"Bearer {access_token}"},
            json={
                "team1": {"score": new_score[0]},
                "team2": {"score": new_score[1]},
            },
        )

        assert response_patch_bet.status_code == HTTPStatus.OK

    response_group_result_response = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_group_result_response.status_code == HTTPStatus.OK
    assert response_group_result_response.json()["result"]["group_rank"] == [
        {
            "team": {
                "id": ANY,
                "code": "FR",
                "description": "France",
                "flag": {"url": ANY},
            },
            "drawn": 1,
            "goals_against": 1,
            "goals_difference": 4,
            "goals_for": 5,
            "lost": 0,
            "played": 2,
            "points": 4,
            "won": 1,
        },
        {
            "team": {
                "id": ANY,
                "code": "IM",
                "description": "Île de Man",
                "flag": {"url": ANY},
            },
            "drawn": 1,
            "goals_against": 1,
            "goals_difference": 1,
            "goals_for": 2,
            "lost": 0,
            "played": 2,
            "points": 4,
            "won": 1,
        },
        {
            "team": {
                "code": "IE",
                "description": "Irlande",
                "id": ANY,
                "flag": {"url": ANY},
            },
            "drawn": 0,
            "goals_against": 7,
            "goals_difference": -5,
            "goals_for": 2,
            "lost": 2,
            "played": 2,
            "points": 0,
            "won": 0,
        },
    ]

    # Error case : retrieve group rank with invalid code
    invalid_group_code = get_random_string(2)

    response_group_rank_with_invalid_code = client.get(
        f"/api/v1/bets/groups/rank/{invalid_group_code}",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_group_rank_with_invalid_code.json() == {
        "description": f"Group not found: {invalid_group_code}",
        "error_code": HTTPStatus.NOT_FOUND,
        "ok": False,
    }

    response_patch_bet = client.patch(
        f"/api/v1/score_bets/{response_all_bets.json()['result']['score_bets'][0]['id']}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={
            "team1": {"score": 6},
            "team2": {"score": 5},
        },
    )

    assert response_patch_bet.status_code == HTTPStatus.OK

    response_group_rank_response = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_group_rank_response.status_code == HTTPStatus.OK
    assert response_group_rank_response.json()["result"]["group_rank"] == [
        {
            "team": {
                "id": ANY,
                "code": ANY,
                "description": "France",
                "flag": ANY,
            },
            "played": 2,
            "won": 1,
            "drawn": 1,
            "lost": 0,
            "goals_for": 6,
            "goals_against": 5,
            "goals_difference": 1,
            "points": 4,
        },
        {
            "team": {
                "id": ANY,
                "code": ANY,
                "description": "Île de Man",
                "flag": ANY,
            },
            "played": 2,
            "won": 1,
            "drawn": 1,
            "lost": 0,
            "goals_for": 2,
            "goals_against": 1,
            "goals_difference": 1,
            "points": 4,
        },
        {
            "team": {
                "id": ANY,
                "code": ANY,
                "description": "Irlande",
                "flag": ANY,
            },
            "played": 2,
            "won": 0,
            "drawn": 0,
            "lost": 2,
            "goals_for": 6,
            "goals_against": 8,
            "goals_difference": -2,
            "points": 0,
        },
    ]

    response_patch_bet = client.patch(
        f"/api/v1/score_bets/{response_all_bets.json()['result']['score_bets'][1]['id']}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={
            "team1": {"score": None},
            "team2": {"score": 5},
        },
    )

    assert response_patch_bet.status_code == HTTPStatus.OK

    response_group_rank_response = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_group_rank_response.json()["result"]["group_rank"] == [
        {
            "team": {
                "id": ANY,
                "code": "FR",
                "description": "France",
                "flag": {"url": ANY},
            },
            "played": 1,
            "won": 1,
            "drawn": 0,
            "lost": 0,
            "goals_for": 6,
            "goals_against": 5,
            "goals_difference": 1,
            "points": 3,
        },
        {
            "team": {
                "id": ANY,
                "code": "IM",
                "description": "Île de Man",
                "flag": {
                    "url": ANY,
                },
            },
            "played": 1,
            "won": 1,
            "drawn": 0,
            "lost": 0,
            "goals_for": 2,
            "goals_against": 1,
            "goals_difference": 1,
            "points": 3,
        },
        {
            "team": {
                "id": ANY,
                "code": "IE",
                "description": "Irlande",
                "flag": {
                    "url": ANY,
                },
            },
            "played": 2,
            "won": 0,
            "drawn": 0,
            "lost": 2,
            "goals_for": 6,
            "goals_against": 8,
            "goals_difference": -2,
            "points": 0,
        },
    ]

    response_patch_bet = client.patch(
        f"/api/v1/score_bets/{response_all_bets.json()['result']['score_bets'][2]['id']}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={
            "team1": {"score": None},
            "team2": {"score": None},
        },
    )

    assert response_patch_bet.status_code == HTTPStatus.OK

    response_group_rank_response = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_group_rank_response.json()["result"]["group_rank"] == [
        {
            "team": {
                "id": ANY,
                "code": "FR",
                "description": "France",
                "flag": {"url": ANY},
            },
            "played": 1,
            "won": 1,
            "drawn": 0,
            "lost": 0,
            "goals_for": 6,
            "goals_against": 5,
            "goals_difference": 1,
            "points": 3,
        },
        {
            "team": {
                "id": ANY,
                "code": "IM",
                "description": "Île de Man",
                "flag": {"url": ANY},
            },
            "played": 0,
            "won": 0,
            "drawn": 0,
            "lost": 0,
            "goals_for": 0,
            "goals_against": 0,
            "goals_difference": 0,
            "points": 0,
        },
        {
            "team": {
                "id": ANY,
                "code": "IE",
                "description": "Irlande",
                "flag": {"url": ANY},
            },
            "played": 1,
            "won": 0,
            "drawn": 0,
            "lost": 1,
            "goals_for": 5,
            "goals_against": 6,
            "goals_difference": -1,
            "points": 0,
        },
    ]

    response_patch_bet = client.patch(
        f"/api/v1/score_bets/{response_all_bets.json()['result']['score_bets'][0]['id']}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={
            "team1": {"score": 1},
            "team2": {"score": None},
        },
    )

    assert response_patch_bet.status_code == HTTPStatus.OK

    response_group_rank_response = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert sorted(
        response_group_rank_response.json()["result"]["group_rank"],
        key=lambda group_position: group_position["team"]["code"],
    ) == sorted(
        [
            {
                "team": {
                    "id": ANY,
                    "code": "FR",
                    "description": "France",
                    "flag": {"url": ANY},
                },
                "won": 0,
                "drawn": 0,
                "lost": 0,
                "goals_for": 0,
                "goals_against": 0,
                "goals_difference": 0,
                "played": 0,
                "points": 0,
            },
            {
                "team": {
                    "id": ANY,
                    "code": "IE",
                    "description": "Irlande",
                    "flag": {"url": ANY},
                },
                "won": 0,
                "drawn": 0,
                "lost": 0,
                "goals_for": 0,
                "goals_against": 0,
                "goals_difference": 0,
                "played": 0,
                "points": 0,
            },
            {
                "team": {
                    "id": ANY,
                    "code": "IM",
                    "description": "Île de Man",
                    "flag": {"url": ANY},
                },
                "won": 0,
                "drawn": 0,
                "lost": 0,
                "goals_for": 0,
                "goals_against": 0,
                "goals_difference": 0,
                "played": 0,
                "points": 0,
            },
        ],
        # I know that the order is not important but I want to be sure that the test is consistent
        key=lambda group_position: group_position["team"]["code"],  # type: ignore[index]
    )

    response_patch_bet = client.patch(
        f"/api/v1/score_bets/{response_all_bets.json()['result']['score_bets'][0]['id']}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={
            "team1": {"score": None},
            "team2": {"score": None},
        },
    )

    assert response_patch_bet.status_code == HTTPStatus.OK

    response_group_rank_response_1 = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert sorted(
        response_group_rank_response_1.json()["result"]["group_rank"],
        key=lambda group_position: group_position["team"]["code"],
    ) == sorted(
        response_group_rank_response.json()["result"]["group_rank"],
        key=lambda group_position: group_position["team"]["code"],
    )


def test_group_rank_team_not_defined(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_group_rank_team_not_defined"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(10),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    access_token = response_signup.json()["result"]["access_token"]

    response_retrieve_bets_group = client.get(
        "/api/v1/bets/groups/A", headers={"Authorization": f"Bearer {access_token}"}
    )

    assert response_retrieve_bets_group.status_code == HTTPStatus.OK

    bet_id = response_retrieve_bets_group.json()["result"]["score_bets"][1]["id"]

    response_modify_score_bet = client.patch(
        f"/api/v1/score_bets/{bet_id}",
        headers={"Authorization": f"Bearer {access_token}"},
        json={"team1": {"score": 2}, "team2": {"score": 0}},
    )

    assert response_modify_score_bet.status_code == HTTPStatus.OK

    response_group_result_response = client.get(
        "/api/v1/bets/groups/rank/A",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_group_result_response.status_code == HTTPStatus.OK



from http import HTTPStatus
from typing import TYPE_CHECKING

import pytest
from starlette.testclient import TestClient

from testing.util import get_random_string
from yak_server.cli.database import (
    RecordDeletionInProductionError,
    TableDropInProductionError,
    create_admin,
    delete_database,
    drop_database,
)

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_create_admin(app_with_valid_jwt_config: "FastAPI", engine_for_test: "Engine") -> None:
    client = TestClient(app_with_valid_jwt_config)

    # Success case : create admin using script and test login is OK
    password_admin = get_random_string(10)

    create_admin(password_admin, engine_for_test)

    response_login = client.post(
        "/api/v1/users/login",
        json={
            "name": "admin",
            "password": password_admin,
        },
    )

    assert response_login.status_code == HTTPStatus.CREATED

    # Success case : if admin already exists in db, it must not fail
    password_admin = get_random_string(10)

    create_admin(password_admin, engine_for_test)


def test_delete_all_records(engine_for_test: "Engine") -> None:
    with pytest.raises(RecordDeletionInProductionError):
        delete_database(engine_for_test, debug=False)


def test_drop_all_tables(engine_for_test: "Engine") -> None:
    with pytest.raises(TableDropInProductionError):
        drop_database(engine_for_test, debug=False)



from http import HTTPStatus
from operator import itemgetter
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_teams(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_teams_v1"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    # Fetch all the teams
    response_get_all_teams = client.get("api/v1/teams")

    assert response_get_all_teams.status_code == HTTPStatus.OK

    excepted_teams = [
        {
            "id": ANY,
            "code": "DE",
            "description": "Allemagne",
            "flag": {"url": ANY},
        },
        {"id": ANY, "code": "GR", "description": "Grèce", "flag": {"url": ANY}},
        {"id": ANY, "code": "GD", "description": "Grenade", "flag": {"url": ANY}},
        {"id": ANY, "code": "JM", "description": "Jamaïque", "flag": {"url": ANY}},
        {"id": ANY, "code": "JO", "description": "Jordanie", "flag": {"url": ANY}},
        {
            "id": ANY,
            "code": "ML",
            "description": "Mali",
            "flag": {"url": ANY},
        },
        {"id": ANY, "code": "NO", "description": "Norvège", "flag": {"url": ANY}},
        {"id": ANY, "code": "UA", "description": "Ukraine", "flag": {"url": ANY}},
    ]

    assert sorted(
        response_get_all_teams.json()["result"]["teams"],
        key=itemgetter("description"),
    ) == sorted(excepted_teams, key=itemgetter("description"))

    # Fetch one team using code
    response_one_team_by_code = client.get(
        "api/v1/teams/DE",
    )

    assert response_one_team_by_code.json()["result"]["team"] == {
        "id": ANY,
        "code": "DE",
        "description": "Allemagne",
        "flag": {"url": ANY},
    }

    # Fetch one team using id
    team_id = response_one_team_by_code.json()["result"]["team"]["id"]

    response_one_team_by_id = client.get(
        f"api/v1/teams/{team_id}",
    )

    assert response_one_team_by_id.json()["result"]["team"] == {
        "id": team_id,
        "code": "DE",
        "description": "Allemagne",
        "flag": {"url": ANY},
    }

    # Try to fetch a team using an invalid team id
    invalid_team_id = f"{team_id}dd"

    response_one_team_with_invalid_id = client.get(
        f"api/v1/teams/{invalid_team_id}",
    )

    assert response_one_team_with_invalid_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.BAD_REQUEST,
        "description": (
            f"Invalid team id: {invalid_team_id}. Retry with a uuid or ISO 3166-1 alpha-2 code"
        ),
    }

    # Try to fetch a team using a non existing team id
    non_existing_team_id = uuid4()

    response_non_existing_team_id = client.get(
        f"api/v1/teams/{non_existing_team_id}",
    )

    assert response_non_existing_team_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Team not found: {non_existing_team_id}",
    }

    # Check flag fetching
    response_retrieve_flag = client.get(
        response_one_team_by_code.json()["result"]["team"]["flag"]["url"],
        follow_redirects=False,
    )

    assert response_retrieve_flag.status_code == HTTPStatus.TEMPORARY_REDIRECT

    # Check flag fetching with invalid team id
    invalid_team_id_with_flag = uuid4()

    response_retrieve_flag_with_invalid_id = client.get(
        f"/api/v1/teams/{invalid_team_id_with_flag}/flag",
    )

    assert response_retrieve_flag_with_invalid_id.status_code == HTTPStatus.NOT_FOUND
    assert response_retrieve_flag_with_invalid_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Team not found: {invalid_team_id_with_flag}",
    }



from typing import Annotated

from fastapi import Depends
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from jwt import ExpiredSignatureError, PyJWTError
from sqlalchemy.orm import Session

from yak_server.database.models import UserModel
from yak_server.helpers.authentication import decode_bearer_token
from yak_server.helpers.database import get_db
from yak_server.helpers.settings import AuthenticationSettings, get_authentication_settings

from .errors import (
    ExpiredToken,
    InvalidToken,
    UnauthorizedAccessToAdminAPI,
    UserNotFound,
)

security = HTTPBearer(auto_error=False)


def user_from_token(db: Session, secret_key: str, access_token: str) -> UserModel:
    try:
        data = decode_bearer_token(access_token, secret_key)
    except ExpiredSignatureError as exc:
        raise ExpiredToken from exc
    except PyJWTError as exc:
        raise InvalidToken from exc

    user = db.query(UserModel).filter_by(id=data["sub"]).first()
    if not user:
        raise UserNotFound(data["sub"])

    return user


def get_current_user(
    access_token: Annotated[HTTPAuthorizationCredentials, Depends(security)],
    db: Annotated[Session, Depends(get_db)],
    auth_settings: Annotated[AuthenticationSettings, Depends(get_authentication_settings)],
) -> UserModel:
    if access_token is None:
        raise InvalidToken

    return user_from_token(db, auth_settings.jwt_secret_key, access_token.credentials)


def get_admin_user(
    access_token: Annotated[HTTPAuthorizationCredentials, Depends(security)],
    db: Annotated[Session, Depends(get_db)],
    auth_settings: Annotated[AuthenticationSettings, Depends(get_authentication_settings)],
) -> UserModel:
    user = get_current_user(access_token, db, auth_settings)

    if user.name != "admin":
        raise UnauthorizedAccessToAdminAPI

    return user



from http import HTTPStatus
from secrets import SystemRandom, randbelow
from typing import TYPE_CHECKING
from uuid import uuid4

import pendulum
from fastapi.testclient import TestClient

from testing.mock import MockLockDatetime, MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database
from yak_server.helpers.settings import get_lock_datetime

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_modify_score_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_modify_bet_v2"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    user_name = get_random_string(10)
    first_name = get_random_string(5)
    last_name = get_random_string(8)
    password = get_random_string(15)

    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED
    authentication_token = response_signup.json()["result"]["access_token"]

    response_get_all_bets = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_get_all_bets.status_code == HTTPStatus.OK

    score_bet_ids = [
        score_bet["id"] for score_bet in response_get_all_bets.json()["result"]["score_bets"]
    ]

    # Success case : check update is OK
    score1 = 4
    score2 = 4

    response_patch_one_bet = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team1": {"score": score1}, "team2": {"score": score2}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_one_bet.status_code == HTTPStatus.OK
    assert response_patch_one_bet.json()["result"]["score_bet"]["team1"]["score"] == score1
    assert response_patch_one_bet.json()["result"]["score_bet"]["team2"]["score"] == score2

    # Success case : check no updates
    response_patch_no_updates = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team1": {"score": score1}, "team2": {"score": score2}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_no_updates.status_code == HTTPStatus.OK
    assert response_patch_no_updates.json()["result"]["score_bet"]["team1"]["score"] == score1
    assert response_patch_no_updates.json()["result"]["score_bet"]["team2"]["score"] == score2

    # Error case : check locked bet
    app_with_valid_jwt_config.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") - pendulum.duration(minutes=10)
    )

    response_locked_bet = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team1": {"score": score1}, "team2": {"score": score2}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_locked_bet.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Cannot modify score bet, lock date is exceeded",
    }

    app_with_valid_jwt_config.dependency_overrides[get_lock_datetime] = MockLockDatetime(
        pendulum.now("UTC") + pendulum.duration(minutes=10)
    )

    # Error case : check bet not found
    non_existing_bet_id = uuid4()

    response_bet_not_found = client.patch(
        f"/api/v1/score_bets/{non_existing_bet_id}",
        json={"team1": {"score": score1}, "team2": {"score": score2}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_bet_not_found.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Bet not found: {non_existing_bet_id}",
    }

    # Error case : check new score negative error
    response_new_score_negative = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[0]}",
        json={"team1": {"score": -1}, "team2": {"score": score2}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_new_score_negative.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNPROCESSABLE_ENTITY,
        "description": [
            {
                "error": "Input should be greater than or equal to 0",
                "field": "body -> team1 -> score",
            }
        ],
    }

    # Patch second bet
    secret_rng = SystemRandom()

    response_patch_second_bet = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[1]}",
        json={"team1": {"score": randbelow(5)}, "team2": {"score": secret_rng.randrange(1, 3)}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_second_bet.status_code == HTTPStatus.OK

    # Patch third bet
    response_patch_third_bet = client.patch(
        f"/api/v1/score_bets/{score_bet_ids[2]}",
        json={"team1": {"score": randbelow(3)}, "team2": {"score": secret_rng.randrange(2, 3)}},
        headers={"Authorization": f"Bearer {authentication_token}"},
    )

    assert response_patch_third_bet.status_code == HTTPStatus.OK



from http import HTTPStatus
from typing import TYPE_CHECKING

from sqlalchemy import TextClause
from sqlalchemy.exc import SQLAlchemyError
from starlette.testclient import TestClient

from yak_server.helpers.database import get_db

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_health_check_ok(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response = client.get("/api/health/")

    assert response.status_code == HTTPStatus.OK
    assert response.json() == {"ok": True}


class FakeDbSession:
    def execute(self, statement: TextClause) -> None:  # noqa: PLR6301
        raise SQLAlchemyError(statement)


def test_health_check_ko(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    app_with_valid_jwt_config.dependency_overrides[get_db] = FakeDbSession

    response = client.get("/api/health/")

    assert response.status_code == HTTPStatus.SERVICE_UNAVAILABLE
    assert response.json() == {"ok": False, "error_code": 503, "description": "Service Unavailable"}



import logging
import traceback
from typing import TYPE_CHECKING, Union
from uuid import UUID

from fastapi import HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from sqlalchemy.exc import SQLAlchemyError
from starlette.exceptions import HTTPException as StarletteHTTPException

from yak_server.helpers.errors import (
    EXPIRED_TOKEN_MESSAGE,
    INVALID_CREDENTIALS_MESSAGE,
    INVALID_TOKEN_MESSAGE,
    LOCKED_BINARY_BET_MESSAGE,
    LOCKED_SCORE_BET_MESSAGE,
    UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
    bet_not_found_message,
    group_not_found_message,
    name_already_exists_message,
    phase_not_found_message,
    rule_not_found_message,
    team_not_found_message,
    user_not_found_message,
)

if TYPE_CHECKING:
    from fastapi import FastAPI

logger = logging.getLogger(__name__)


class InvalidCredentials(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=INVALID_CREDENTIALS_MESSAGE,
        )


class NameAlreadyExists(HTTPException):
    def __init__(self, name: str) -> None:
        super().__init__(
            status_code=status.HTTP_409_CONFLICT,
            detail=name_already_exists_message(name),
        )


class InvalidRefreshToken(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid refresh token",
        )


class ExpiredRefreshToken(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Refresh token expired",
        )


class BetNotFound(HTTPException):
    def __init__(self, bet_id: UUID) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=bet_not_found_message(bet_id),
        )


class UnsatisfiedPasswordRequirements(HTTPException):
    def __init__(self, detail: str) -> None:
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Unsatisfied password requirements. {detail}",
        )


class UserNotFound(HTTPException):
    def __init__(self, user_id: UUID) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=user_not_found_message(user_id),
        )


class UnauthorizedAccessToAdminAPI(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
        )


class NoAdminUser(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="No admin user found",
        )


class InvalidTeamId(HTTPException):
    def __init__(self, team_id: str) -> None:
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid team id: {team_id}. Retry with a uuid or ISO 3166-1 alpha-2 code",
        )


class TeamNotFound(HTTPException):
    def __init__(self, team_id: Union[str, UUID]) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=team_not_found_message(team_id),
        )


class LockedScoreBet(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=LOCKED_SCORE_BET_MESSAGE)


class LockedBinaryBet(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=LOCKED_BINARY_BET_MESSAGE)


class NoResultsForAdminUser(HTTPException):
    def __init__(self) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="No results for admin user",
        )


class GroupNotFound(HTTPException):
    def __init__(self, group_id: Union[str, UUID]) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=group_not_found_message(group_id),
        )


class PhaseNotFound(HTTPException):
    def __init__(self, phase_id: Union[str, UUID]) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=phase_not_found_message(phase_id),
        )


class InvalidToken(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=INVALID_TOKEN_MESSAGE)


class ExpiredToken(HTTPException):
    def __init__(self) -> None:
        super().__init__(status_code=status.HTTP_401_UNAUTHORIZED, detail=EXPIRED_TOKEN_MESSAGE)


class RuleNotFound(HTTPException):
    def __init__(self, rule_id: UUID) -> None:
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=rule_not_found_message(rule_id),
        )


def log_traceback(exception: Exception) -> None:
    """
    Log the traceback of an exception.
    """
    for frame in traceback.format_list(traceback.extract_tb(exception.__traceback__)):
        logger.info(frame)

    if exception.__cause__:
        for frame in traceback.format_list(traceback.extract_tb(exception.__cause__.__traceback__)):
            logger.info(frame)


def set_exception_handler(app: "FastAPI") -> None:
    @app.exception_handler(StarletteHTTPException)
    def http_exception_handler(_: Request, http_exception: StarletteHTTPException) -> JSONResponse:
        log_traceback(http_exception)

        logger.info(
            f"An expected exception occurs: {type(http_exception).__name__} {http_exception}"
        )

        return JSONResponse(
            status_code=http_exception.status_code,
            content={
                "ok": False,
                "error_code": http_exception.status_code,
                "description": http_exception.detail,
            },
        )

    @app.exception_handler(Exception)
    def handle_exception(_: Request, exception: Exception) -> JSONResponse:  # pragma: no cover
        # Return JSON instead of HTML for generic errors.
        log_traceback(exception)

        logger.error(f"An unexpected exception occurs: {type(exception).__name__} {exception}")

        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={
                "ok": False,
                "error_code": status.HTTP_500_INTERNAL_SERVER_ERROR,
                "description": (
                    f"{type(exception).__name__}: {exception!s}"
                    if app.debug
                    else "Unexpected error"
                ),
            },
        )

    @app.exception_handler(RequestValidationError)
    def request_validator_error_handler(
        _: Request,
        request_validator_error: RequestValidationError,
    ) -> JSONResponse:
        log_traceback(request_validator_error)

        logger.info(
            "An validation error occurs: "
            f"{type(request_validator_error).__name__} {request_validator_error}"
        )

        errors = []

        for err in request_validator_error.errors():
            loc = " -> ".join(str(loc) for loc in err["loc"])
            msg = err["msg"]
            errors.append({"field": loc, "error": msg})

        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content={
                "ok": False,
                "error_code": status.HTTP_422_UNPROCESSABLE_ENTITY,
                "description": errors,
            },
        )

    @app.exception_handler(SQLAlchemyError)
    def sql_alchemy_error_handler(_: Request, sql_alchemy_error: SQLAlchemyError) -> JSONResponse:
        log_traceback(sql_alchemy_error)

        logger.error(f"Database error: {type(sql_alchemy_error).__name__} {sql_alchemy_error}")

        return JSONResponse(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
            content={
                "ok": False,
                "error_code": status.HTTP_503_SERVICE_UNAVAILABLE,
                "description": "Service Unavailable",
            },
        )



from functools import cache
from typing import Annotated

import pendulum
from pydantic import HttpUrl, PlainValidator
from pydantic_settings import BaseSettings, SettingsConfigDict

from .rules import Rules

PendulumDateTime = Annotated[pendulum.DateTime, PlainValidator(pendulum.parse)]


class Settings(BaseSettings):
    competition: str
    data_folder: str
    rules: Rules
    official_results_url: HttpUrl

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_settings() -> Settings:
    return Settings()


class LockDatetimeSettings(BaseSettings):
    lock_datetime: PendulumDateTime

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_lock_datetime() -> pendulum.DateTime:
    return LockDatetimeSettings().lock_datetime  # pragma: no cover


class AuthenticationSettings(BaseSettings):
    jwt_secret_key: str
    jwt_refresh_secret_key: str
    jwt_expiration_time: int
    jwt_refresh_expiration_time: int

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_authentication_settings() -> AuthenticationSettings:
    return AuthenticationSettings()  # pragma: no cover



from typing import TYPE_CHECKING

from pydantic import BaseModel, NonNegativeInt

from .teams import TeamOut

if TYPE_CHECKING:
    from yak_server.database.models import GroupPositionModel
    from yak_server.helpers.language import Lang


class GroupPositionOut(BaseModel):
    team: TeamOut
    played: NonNegativeInt
    won: NonNegativeInt
    drawn: NonNegativeInt
    lost: NonNegativeInt
    goals_for: NonNegativeInt
    goals_against: NonNegativeInt
    goals_difference: int
    points: NonNegativeInt

    @classmethod
    def from_instance(
        cls,
        group_position: "GroupPositionModel",
        *,
        lang: "Lang",
    ) -> "GroupPositionOut":
        return cls(
            team=TeamOut.from_instance(group_position.team, lang=lang),
            played=group_position.played,
            won=group_position.won,
            drawn=group_position.drawn,
            lost=group_position.lost,
            goals_for=group_position.goals_for,
            goals_against=group_position.goals_against,
            goals_difference=group_position.goals_difference,
            points=group_position.points,
        )



import logging
from typing import Annotated

import pendulum
from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, selectinload

from yak_server.database.models import GroupModel, MatchModel, ScoreBetModel, UserModel
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.database import get_db
from yak_server.helpers.group_position import set_recomputation_flag
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang, get_language_description
from yak_server.helpers.logging_helpers import modify_score_bet_successfully
from yak_server.helpers.settings import get_lock_datetime
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.errors import BetNotFound, LockedScoreBet, TeamNotFound
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut
from yak_server.v1.models.groups import GroupOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.score_bets import ModifyScoreBetIn, ScoreBetOut, ScoreBetResponse
from yak_server.v1.models.teams import FlagOut, TeamWithScoreOut

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/score_bets", tags=["score_bets"])


def send_response(
    score_bet: ScoreBetModel,
    *,
    locked: bool,
    lang: Lang,
) -> GenericOut[ScoreBetResponse]:
    return GenericOut(
        result=ScoreBetResponse(
            phase=PhaseOut.from_instance(score_bet.match.group.phase, lang=lang),
            group=GroupOut.from_instance(score_bet.match.group, lang=lang),
            score_bet=ScoreBetOut(
                id=score_bet.id,
                locked=locked,
                team1=(
                    TeamWithScoreOut(
                        id=score_bet.match.team1.id,
                        code=score_bet.match.team1.code,
                        description=get_language_description(score_bet.match.team1, lang),
                        flag=FlagOut(url=score_bet.match.team1.flag_url),
                        score=score_bet.score1,
                    )
                    if score_bet.match.team1 is not None
                    else None
                ),
                team2=(
                    TeamWithScoreOut(
                        id=score_bet.match.team2.id,
                        code=score_bet.match.team2.code,
                        description=get_language_description(score_bet.match.team2, lang),
                        flag=FlagOut(url=score_bet.match.team2.flag_url),
                        score=score_bet.score2,
                    )
                    if score_bet.match.team2 is not None
                    else None
                ),
            ),
        ),
    )


@router.get(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_ENTITY: {"model": ValidationErrorOut},
    },
)
def retrieve_score_bet_by_id(
    bet_id: UUID4,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(get_current_user)],
    lock_datetime: Annotated[pendulum.DateTime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[ScoreBetResponse]:
    score_bet = (
        db.query(ScoreBetModel)
        .options(
            selectinload(ScoreBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team1),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team2),
        )
        .join(ScoreBetModel.match)
        .where(MatchModel.user_id == user.id, ScoreBetModel.id == bet_id)
        .first()
    )

    if not score_bet:
        raise BetNotFound(bet_id)

    return send_response(score_bet, locked=is_locked(user.name, lock_datetime), lang=lang)


@router.patch(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_ENTITY: {"model": ValidationErrorOut},
    },
)
def modify_score_bet(
    bet_id: UUID4,
    modify_score_bet_in: ModifyScoreBetIn,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(get_current_user)],
    lock_datetime: Annotated[pendulum.DateTime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[ScoreBetResponse]:
    if is_locked(user.name, lock_datetime):
        raise LockedScoreBet

    score_bet = (
        db.query(ScoreBetModel)
        .options(
            selectinload(ScoreBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team1),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team2),
        )
        .join(ScoreBetModel.match)
        .where(MatchModel.user_id == user.id, ScoreBetModel.id == bet_id)
        .with_for_update()
        .first()
    )

    if not score_bet:
        raise BetNotFound(bet_id)

    logger.info(
        modify_score_bet_successfully(
            user.name,
            score_bet,
            modify_score_bet_in.team1.score if modify_score_bet_in.team1 else None,
            modify_score_bet_in.team2.score if modify_score_bet_in.team2 else None,
        ),
    )

    if modify_score_bet_in.team1 is not None:
        if "id" in modify_score_bet_in.team1.model_fields_set:
            score_bet.match.team1_id = modify_score_bet_in.team1.id

            try:
                db.flush()
            except IntegrityError as integrity_error:
                db.rollback()
                # team1.id is not None due to: "id" in modify_score_bet_in.team1.model_fields_set
                # being true
                raise TeamNotFound(modify_score_bet_in.team1.id) from integrity_error  # type: ignore[arg-type]

        if "score" in modify_score_bet_in.team1.model_fields_set:
            score_bet.score1 = modify_score_bet_in.team1.score

    if modify_score_bet_in.team2 is not None:
        if "id" in modify_score_bet_in.team2.model_fields_set:
            score_bet.match.team2_id = modify_score_bet_in.team2.id

            try:
                db.flush()
            except IntegrityError as integrity_error:
                db.rollback()
                # team2.id is not None due to: "id" in modify_score_bet_in.team2.model_fields_set
                # being true
                raise TeamNotFound(modify_score_bet_in.team2.id) from integrity_error  # type: ignore[arg-type]

        if "score" in modify_score_bet_in.team2.model_fields_set:
            score_bet.score2 = modify_score_bet_in.team2.score

    set_recomputation_flag(db, score_bet.match.team1_id, user.id)
    set_recomputation_flag(db, score_bet.match.team2_id, user.id)

    db.commit()
    db.refresh(score_bet)

    return send_response(score_bet, locked=is_locked(user.name, lock_datetime), lang=lang)



from collections.abc import Generator

from sqlalchemy.orm import Session

from yak_server.database import build_engine, build_local_session_maker


def get_db() -> Generator[Session, None, None]:
    engine = build_engine()

    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        yield db



import logging
from typing import Annotated

import pendulum
from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, selectinload

from yak_server.database.models import BinaryBetModel, GroupModel, MatchModel, UserModel
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.database import get_db
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang, get_language_description
from yak_server.helpers.logging_helpers import modify_binary_bet_successfully
from yak_server.helpers.settings import get_lock_datetime
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.errors import BetNotFound, LockedBinaryBet, TeamNotFound
from yak_server.v1.models.binary_bets import BinaryBetOut, BinaryBetResponse, ModifyBinaryBetIn
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut
from yak_server.v1.models.groups import GroupOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.teams import FlagOut, TeamWithWonOut

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/binary_bets", tags=["binary_bets"])


def send_response(
    binary_bet: BinaryBetModel,
    *,
    locked: bool,
    lang: Lang,
) -> GenericOut[BinaryBetResponse]:
    return GenericOut(
        result=BinaryBetResponse(
            phase=PhaseOut.from_instance(binary_bet.match.group.phase, lang=lang),
            group=GroupOut.from_instance(binary_bet.match.group, lang=lang),
            binary_bet=BinaryBetOut(
                id=binary_bet.id,
                locked=locked,
                team1=(
                    TeamWithWonOut(
                        id=binary_bet.match.team1.id,
                        code=binary_bet.match.team1.code,
                        description=get_language_description(binary_bet.match.team1, lang),
                        flag=FlagOut(url=binary_bet.match.team1.flag_url),
                        won=binary_bet.bet_from_is_one_won()[0],
                    )
                    if binary_bet.match.team1 is not None
                    else None
                ),
                team2=(
                    TeamWithWonOut(
                        id=binary_bet.match.team2.id,
                        code=binary_bet.match.team2.code,
                        description=get_language_description(binary_bet.match.team2, lang),
                        flag=FlagOut(url=binary_bet.match.team2.flag_url),
                        won=binary_bet.bet_from_is_one_won()[1],
                    )
                    if binary_bet.match.team2 is not None
                    else None
                ),
            ),
        ),
    )


@router.get(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_ENTITY: {"model": ValidationErrorOut},
    },
)
def retrieve_binary_bet_by_id(
    bet_id: UUID4,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(get_current_user)],
    lock_datetime: Annotated[pendulum.DateTime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[BinaryBetResponse]:
    binary_bet = (
        db.query(BinaryBetModel)
        .options(
            selectinload(BinaryBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.group),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team1),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team2),
        )
        .join(BinaryBetModel.match)
        .where(MatchModel.user_id == user.id, BinaryBetModel.id == bet_id)
        .first()
    )

    if not binary_bet:
        raise BetNotFound(bet_id)

    return send_response(binary_bet, locked=is_locked(user.name, lock_datetime), lang=lang)


@router.patch(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_ENTITY: {"model": ValidationErrorOut},
    },
)
def modify_binary_bet_by_id(
    bet_id: UUID4,
    modify_binary_bet_in: ModifyBinaryBetIn,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(get_current_user)],
    lock_datetime: Annotated[pendulum.DateTime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[BinaryBetResponse]:
    if is_locked(user.name, lock_datetime):
        raise LockedBinaryBet

    binary_bet = (
        db.query(BinaryBetModel)
        .options(
            selectinload(BinaryBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.group),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team1),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team2),
        )
        .join(BinaryBetModel.match)
        .where(MatchModel.user_id == user.id, BinaryBetModel.id == bet_id)
        .first()
    )

    if not binary_bet:
        raise BetNotFound(bet_id)

    logger.info(
        modify_binary_bet_successfully(
            user.name, binary_bet, new_is_one_won=modify_binary_bet_in.is_one_won
        ),
    )

    if "is_one_won" in modify_binary_bet_in.model_fields_set:
        binary_bet.is_one_won = modify_binary_bet_in.is_one_won

    if (
        modify_binary_bet_in.team1 is not None
        and "id" in modify_binary_bet_in.team1.model_fields_set
    ):
        binary_bet.match.team1_id = modify_binary_bet_in.team1.id

        try:
            db.flush()
        except IntegrityError as integrity_error:
            db.rollback()
            # team1.id is not None due to: "id" in modify_binary_bet_in.team1.model_fields_set
            # being true
            raise TeamNotFound(modify_binary_bet_in.team1.id) from integrity_error  # type: ignore[arg-type]

    if (
        modify_binary_bet_in.team2 is not None
        and "id" in modify_binary_bet_in.team2.model_fields_set
    ):
        binary_bet.match.team2_id = modify_binary_bet_in.team2.id

        try:
            db.flush()
        except IntegrityError as integrity_error:
            db.rollback()
            # team2.id is not None due to: "id" in modify_binary_bet_in.team2.model_fields_set
            # being true
            raise TeamNotFound(modify_binary_bet_in.team2.id) from integrity_error  # type: ignore[arg-type]

    db.commit()
    db.refresh(binary_bet)

    return send_response(binary_bet, locked=is_locked(user.name, lock_datetime), lang=lang)



from typing import Any
from uuid import UUID

import jwt
import pendulum
from sqlalchemy.orm import Session, selectinload

from yak_server.database.models import (
    MatchModel,
    MatchReferenceModel,
    ScoreBetModel,
    UserModel,
)

from .errors import name_already_exists_message
from .group_position import create_group_position
from .password_validator import validate_password


def encode_bearer_token(
    sub: UUID,
    expiration_time: pendulum.Duration,
    secret_key: str,
) -> str:
    return jwt.encode(
        {
            "sub": str(sub),
            "nbf": pendulum.now("UTC") - pendulum.duration(seconds=3),
            "exp": pendulum.now("UTC") + expiration_time,
        },
        secret_key,
        algorithm="HS512",
    )


def decode_bearer_token(token: str, secret_key: str) -> Any:  # noqa: ANN401
    return jwt.decode(token, secret_key, algorithms=["HS512"])


class NameAlreadyExistsError(Exception):
    def __init__(self, name: str) -> None:
        super().__init__(name_already_exists_message(name))


def signup_user(
    db: Session, name: str, first_name: str, last_name: str, password: str
) -> UserModel:
    # Check existing user in db
    existing_user = db.query(UserModel).filter_by(name=name).first()
    if existing_user:
        raise NameAlreadyExistsError(name)

    # Validate password
    validate_password(password)

    # Initialize user and integrate in db
    user = UserModel(name, first_name, last_name, password)
    db.add(user)
    db.flush()

    # Initialize matches and bets and integrate in db
    for match_reference in db.query(MatchReferenceModel).all():
        match = MatchModel(
            team1_id=match_reference.team1_id,
            team2_id=match_reference.team2_id,
            index=match_reference.index,
            group_id=match_reference.group_id,
            user_id=user.id,
        )
        db.add(match)
        db.flush()

        db.add(match_reference.bet_type_from_match.value(match_id=match.id))
        db.flush()

    # Create group position records
    db.add_all(
        create_group_position(
            db.query(ScoreBetModel)
            .options(selectinload(ScoreBetModel.match))
            .join(ScoreBetModel.match)
            .filter_by(user_id=user.id)
        )
    )
    db.commit()

    return user



import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from uuid import UUID

from sqlalchemy import update
from sqlalchemy.dialects.postgresql import insert

from yak_server.database import build_local_session_maker
from yak_server.database.models import (
    Base,
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    MatchReferenceModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.authentication import NameAlreadyExistsError, signup_user
from yak_server.helpers.rules.compute_points import compute_points as compute_points_func
from yak_server.helpers.settings import get_settings
from yak_server.v1.helpers.errors import NoAdminUser

try:
    import alembic
except ImportError:  # pragma: no cover
    # Very common pattern for optional dependency imports
    alembic = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine
    from sqlalchemy.orm import Session

logger = logging.getLogger(__name__)


class RecordDeletionInProductionError(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to delete records in production using script. DO NOT DO IT.")


class TableDropInProductionError(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to drop database tables in production using script. DO NOT DO IT.")


def create_database(engine: "Engine") -> None:
    Base.metadata.create_all(bind=engine)


def create_admin(password: str, engine: "Engine") -> None:
    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        try:
            _ = signup_user(
                db, name="admin", first_name="admin", last_name="admin", password=password
            )
        except NameAlreadyExistsError:
            print("Admin already exists")


class MissingPhaseDuringInitError(Exception):
    def __init__(self, phase_code: str) -> None:
        super().__init__(
            f"Error during database initialization: phase_code={phase_code} not found."
        )


class MissingTeamDuringInitError(Exception):
    def __init__(self, team_code: str) -> None:
        super().__init__(f"Error during database initialization: team_code={team_code} not found.")


class MissingGroupDuringInitError(Exception):
    def __init__(self, group_code: str) -> None:
        super().__init__(
            f"Error during database initialization: group_code={group_code} not found."
        )


def fetch_team_id(team_code: Optional[str], db: "Session") -> Optional[UUID]:
    if team_code is None:
        return None

    team = db.query(TeamModel).filter_by(code=team_code).first()

    if team is None:
        raise MissingTeamDuringInitError(team_code)

    return team.id


def fetch_group_id(group_code: str, db: "Session") -> UUID:
    group = db.query(GroupModel).filter_by(code=group_code).first()

    if group is None:
        raise MissingGroupDuringInitError(group_code)

    return group.id


def initialize_database(engine: "Engine", app: "FastAPI") -> None:
    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        data_folder = get_settings().data_folder

        phases = json.loads(Path(data_folder, "phases.json").read_text(encoding="utf-8"))

        for phase in phases:
            db.execute(
                insert(PhaseModel).values(**phase).on_conflict_do_nothing(index_elements=["code"])
            )

        db.flush()

        groups = json.loads(Path(data_folder, "groups.json").read_text(encoding="utf-8"))

        for group in groups:
            phase_code = group.pop("phase_code")

            phase = db.query(PhaseModel).filter_by(code=phase_code).first()

            if phase is None:
                raise MissingPhaseDuringInitError(phase_code)

            group["phase_id"] = phase.id

        for group in groups:
            db.execute(
                insert(GroupModel).values(**group).on_conflict_do_nothing(index_elements=["code"])
            )

        db.flush()

        teams = json.loads(Path(data_folder, "teams.json").read_text(encoding="utf-8"))

        for team in teams:
            team["flag_url"] = ""

            db.execute(
                insert(TeamModel).values(**team).on_conflict_do_nothing(index_elements=["code"])
            )

            db.flush()

            team_instance = db.query(TeamModel).filter_by(code=team["code"]).first()

            if team_instance is None:  # pragma: no cover
                raise MissingTeamDuringInitError(team_code=team["code"])

            update_flag_url_stmt = (
                update(TeamModel)
                .where(TeamModel.code == team["code"])
                .values(
                    flag_url=app.url_path_for("retrieve_team_flag_by_id", team_id=team_instance.id)
                )
            )

            db.execute(update_flag_url_stmt)
            db.flush()

        matches = json.loads(Path(data_folder, "matches.json").read_text(encoding="utf-8"))

        for match in matches:
            match["team1_id"] = fetch_team_id(match.pop("team1_code"), db)
            match["team2_id"] = fetch_team_id(match.pop("team2_code"), db)
            match["group_id"] = fetch_group_id(match.pop("group_code"), db)

        for match in matches:
            db.execute(
                insert(MatchReferenceModel)
                .values(**match)
                .on_conflict_do_nothing(index_elements=["group_id", "index"])
            )

        db.flush()

        db.commit()


def delete_database(engine: "Engine", *, debug: bool) -> None:
    if debug is False:
        raise RecordDeletionInProductionError

    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        db.query(GroupPositionModel).delete()
        db.query(ScoreBetModel).delete()
        db.query(BinaryBetModel).delete()
        db.query(MatchReferenceModel).delete()
        db.query(MatchModel).delete()
        db.query(UserModel).delete()
        db.query(GroupModel).delete()
        db.query(PhaseModel).delete()
        db.query(TeamModel).delete()
        db.commit()


def drop_database(engine: "Engine", *, debug: bool) -> None:
    if debug is False:
        raise TableDropInProductionError

    Base.metadata.drop_all(bind=engine)


def print_export_command(alembic_ini_path: Path) -> None:
    print(f"export ALEMBIC_CONFIG={alembic_ini_path}")


def setup_migration(*, short: bool = False) -> None:
    alembic_ini_path = (Path(__file__).parents[2] / "alembic.ini").resolve()

    if not alembic_ini_path.exists():
        alembic_ini_path = (Path(__file__).parents[3] / "alembic.ini").resolve()

    if short is True:
        print_export_command(alembic_ini_path)
    else:
        print(
            "To be able to run the database migration scripts, "
            "you need to run the following command:",
        )
        print_export_command(alembic_ini_path)
        print()
        print(
            "Follow this link for more information: "
            "https://alembic.sqlalchemy.org/en/latest/tutorial.html#editing-the-ini-file",
        )

        if alembic is None:
            print()
            print(
                "To enable migration using alembic, please run: "
                "uv pip install yak-server[db_migration]"
            )


class ComputePointsRuleNotDefinedError(Exception):
    def __init__(self) -> None:
        super().__init__("Compute points rule is not defined.")


def compute_score_board(engine: "Engine") -> None:
    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        admin = db.query(UserModel).filter_by(name="admin").first()

        if admin is None:
            raise NoAdminUser

        rule_config = get_settings().rules.compute_points

        if rule_config is None:
            raise ComputePointsRuleNotDefinedError

        compute_points_func(db, admin, rule_config)
